import datetime
import warnings

import dateutil
#import pyspark.sql.functions as sf
#from pyspark.sql.types import StringType
from unidecode import unidecode

from colabfit import MAX_STRING_LENGTH
#from colabfit.tools.schema import dataset_schema
from colabfit.tools.utilities import ELEMENT_MAP, _empty_dict_from_schema, _hash

import numpy as np


class Dataset:
    """
    A dataset defines a group of configuration sets and computed properties, and
    aggregates information about those configuration sets and properties.

    Attributes:

        configuration_set_ids (list):
            A list of attached configuration sets

        property_ids (list):
            A list of attached properties

        name (str):
            The name of the dataset

        authors (list or str or None):
            The names of the authors of the dataset.

        links (list or str or None):
            External links (e.g., journal articles, Git repositories, ...)
            to be associated with the dataset.

        description (str or None):
            A human-readable description of the dataset.

        aggregated_info (dict):
            A dictionary of information that was aggregated rom all of the
            attached configuration sets and properties. Contains the following
            information:

                From the configuration sets:
                    nconfigurations
                    nsites
                    nelements
                    chemical_systems
                    elements
                    individual_elements_ratios
                    total_elements_ratios
                    configuration_labels
                    configuration_labels_counts
                    chemical_formula_reduced
                    chemical_formula_anonymous
                    chemical_formula_hill
                    nperiodic_dimensions
                    dimension_types

                From the properties:
                    property_types
                    property_fields
                    methods
                    methods_counts
                    property_labels
                    property_labels_counts

        data_license (str):
            License associated with the Dataset's data
    """

    def __init__(
        self,
        name: str,
        authors: list[str],
        publication_link: str,
        data_link: str,
        description: str,
        config_df,
        prop_df,
        property_definitions,
        other_links: list[str] = None,
        dataset_id: str = None,
        labels: list[str] = None,
        doi: str = None,
        configuration_set_ids: list[str] = [],
        data_license: str = "CC-BY-ND-4.0",
        publication_year: str = None,
        use_pg: bool = False,
    ):
        for auth in authors:
            if not "".join(auth.split(" ")[-1].replace("-", "")).isalpha():
                raise RuntimeError(
                    f"Bad author name '{auth}'. Author names "
                    "can only contain [a-z][A-Z]"
                )

        self.property_definitions = property_definitions
        self.name = name
        self.authors = authors
        self.publication_link = publication_link
        self.data_link = data_link
        self.other_links = other_links
        self.description = description
        self.data_license = data_license
        self.dataset_id = dataset_id
        self.doi = doi
        self.publication_year = publication_year
        self.configuration_set_ids = configuration_set_ids
        if self.configuration_set_ids is None:
            self.configuration_set_ids = []
        if use_pg:
            self.spark_row = self._from_pg(configs=config_df, props=prop_df)
        else:
            self.spark_row = self.to_spark_row(config_df=config_df, prop_df=prop_df)

        self.spark_row["id"] = self.dataset_id
        # if dataset_id is None:
        #     raise ValueError("Dataset ID must be provided")
        id_prefix = "__".join(
            [
                self.name,
                "-".join([unidecode(auth.split()[-1]) for auth in authors]),
            ]
        )
        if len(id_prefix) > (MAX_STRING_LENGTH - len(dataset_id) - 2):
            id_prefix = id_prefix[: MAX_STRING_LENGTH - len(dataset_id) - 2]
            warnings.warn(f"ID prefix is too long. Clipping to {id_prefix}")
        extended_id = f"{id_prefix}__{dataset_id}"
        self.spark_row["extended_id"] = extended_id
        self._hash = _hash(self.spark_row, ["extended_id"])
        self.spark_row["hash"] = str(self._hash)
        self.spark_row["labels"] = labels
        
    # aggregate stuff
    def _from_pg(self, configs, props):
        row_dict = {}
        row_dict["last_modified"] = dateutil.parser.parse(
            datetime.datetime.now(tz=datetime.timezone.utc).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            ))
        row_dict["nconfiguration_sets"] = len(self.configuration_set_ids)
        row_dict['nconfigurations'] = len(props)
        row_dict['nproperty_objects'] = len(props)
        nsites = 0
        nperiodic_dimensions = set()
        dimension_types = set()
        element_dict = {} 

        # TODO: Streamline this
        for p in props:
            co_id = p['configuration_id']
            

            for c in configs:
                if c['id'] == co_id:
                    nsites += c['nsites']
                    for e in c['atomic_numbers']:
                       e = ELEMENT_MAP[e] 
                       if e in element_dict:
                            element_dict[e] += 1
                       else:
                            element_dict[e] = 1
                    nperiodic_dimensions.add(c['nperiodic_dimensions'])
                    dimension_types.add(str(c['dimension_types']))

        sorted_elements = sorted(list(element_dict.keys()))

        row_dict['nsites'] = nsites
        row_dict['nelements'] = len(sorted_elements)
        row_dict['elements'] = sorted_elements
        row_dict['total_elements_ratio'] = [element_dict[e] / nsites for e in sorted_elements]
        row_dict['nperiodic_dimensions'] = list(nperiodic_dimensions)
        row_dict['dimension_types'] = list(dimension_types)

        forces = 0
        stress = 0
        energy = 0
        energies = []
        
        # TODO: Below
        prop_def_map = {}
        prop_counts = {}
        # find available props from PD table
        for pd in self.property_definitions:
            # grab first property key name
            for k,v in pd.items():
                if k not in ['property-id', 'property-name', 'property-title', 'property-description']:
                    prop_def_map[f"{pd['property-name'].replace('-','_')}_{k.replace('-','_')}"] = pd['property-name']
                    prop_counts[f"{pd['property-name'].replace('-','_')}_{k.replace('-','_')}"] = 0
                    break

        for p in props:
            for k in p.keys():
                if str(k) in prop_counts:
                    if p[str(k)] is not None:
                        prop_counts[str(k)] += 1
            if p["atomic_forces_forces"] is not None:
                forces += 1
            if p["cauchy_stress_stress"] is not None:
                stress += 1
            if p["energy_energy"] is not None:
                energy += 1
                energies.append(p["energy_energy"])
        # TODO: iterate over counts if >0 add to available_properties
        available_props = []
        for k,v in prop_counts.items():
            if v > 0:
                available_props.append(prop_def_map[k])
        print (available_props)
        row_dict['available_properties'] = available_props
        row_dict['energy_mean'] = np.mean(energies)
        row_dict['energy_variance'] = np.var(energies)
        row_dict['atomic_forces_count'] = forces
        row_dict['cauchy_stress_count'] = stress
        row_dict['energy_count'] = energy
        row_dict["authors"] = self.authors
        row_dict["description"] = self.description
        row_dict["license"] = self.data_license
        row_dict["links"] = str(
            {
                "source-publication": self.publication_link,
                "source-data": self.data_link,
                "other": self.other_links,
            }
        )
        row_dict["name"] = self.name
        row_dict["publication_year"] = self.publication_year
        row_dict["doi"] = self.doi

        return row_dict

    def to_spark_row(self, config_df, prop_df):
        """"""
        row_dict = _empty_dict_from_schema(dataset_schema)
        row_dict["last_modified"] = dateutil.parser.parse(
            datetime.datetime.now(tz=datetime.timezone.utc).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )
        )
        row_dict["nconfiguration_sets"] = len(self.configuration_set_ids)
        config_df = (
            config_df.withColumnRenamed("id", "configuration_id")
            .withColumnRenamed("hash", "config_hash")
            .select(
                "configuration_id",
                "elements",
                "atomic_numbers",
                "nsites",
                "nperiodic_dimensions",
                "dimension_types",
                # "labels",
            )
        )
        nproperty_objects = prop_df.count()
        # co_po_df = prop_df.select(
        #     "configuration_id",
        #     # "multiplicity",
        #     "atomization_energy",
        #     "atomic_forces_00",
        #     "adsorption_energy",
        #     "electronic_band_gap",
        #     "cauchy_stress",
        #     "formation_energy",
        #     "energy",
        # ).join(config_df, on="configuration_id", how="inner")
        # print(co_po_df.columns)
        # print(co_po_df.count())
        # print(co_po_df.first())
        # co_po_df = co_po_df.withColumn(
        #     "nsites_multiple", sf.col("nsites") * sf.col("multiplicity")
        # )
        # row_dict["nsites"] = co_po_df.agg({"nsites_multiple": "sum"}).first()[0]
        row_dict["nsites"] = config_df.agg({"nsites": "sum"}).first()[0]
        row_dict["nproperty_objects"] = nproperty_objects
        row_dict["elements"] = sorted(
            config_df.withColumn("exploded_elements", sf.explode("elements"))
            .agg(sf.collect_set("exploded_elements").alias("exploded_elements"))
            .select("exploded_elements")
            .take(1)[0][0]
        )
        row_dict["nelements"] = len(row_dict["elements"])
        atomic_ratios_df = (
            config_df.select("atomic_numbers")
            #     co_po_df.select("atomic_numbers", "multiplicity")
            #     .withColumn(
            #         "repeated_numbers",
            #         sf.expr(
            #            "transform(atomic_numbers, x -> array_repeat(x, multiplicity))"
            #         ),
            #     )
            # .withColumn("single_element", sf.explode(sf.flatten("repeated_numbers")))
            .withColumn("single_element", sf.explode("atomic_numbers"))
        )
        total_elements = atomic_ratios_df.count()
        assert total_elements == row_dict["nsites"]
        atomic_ratios_df = atomic_ratios_df.groupBy("single_element").count()
        atomic_ratios_df = atomic_ratios_df.withColumn(
            "ratio", sf.col("count") / total_elements
        )

        atomic_ratios_coll = (
            atomic_ratios_df.withColumn(
                "element",
                sf.udf(lambda x: ELEMENT_MAP[x], StringType())(
                    sf.col("single_element")
                ),
            )
            .select("element", "ratio")
            .collect()
        )
        row_dict["total_elements_ratios"] = [
            x[1] for x in sorted(atomic_ratios_coll, key=lambda x: x["element"])
        ]

        row_dict["nperiodic_dimensions"] = config_df.agg(
            sf.collect_set("nperiodic_dimensions")
        ).collect()[0][0]

        row_dict["dimension_types"] = config_df.agg(
            sf.collect_set("dimension_types")
        ).collect()[0][0]

        for prop in [
            "atomization_energy",
            "atomic_forces_00",
            "adsorption_energy",
            "electronic_band_gap",
            "cauchy_stress",
            "formation_energy",
            "energy",
        ]:
            row_dict[f"{prop}_count"] = (
                prop_df.select(prop).where(f"{prop} is not null").count()
            )
        row_dict["atomic_forces_count"] = row_dict.pop("atomic_forces_00_count")

        prop = "energy"
        row_dict[f"{prop}_variance"] = (
            prop_df.select(prop).where(f"{prop} is not null").agg(sf.variance(prop))
        ).first()[0]
        row_dict[f"{prop}_mean"] = (
            prop_df.select(prop).where(f"{prop} is not null").agg(sf.mean(prop))
        ).first()[0]

        row_dict["nconfigurations"] = prop_df.count()
        row_dict["authors"] = self.authors
        row_dict["description"] = self.description
        row_dict["license"] = self.data_license
        row_dict["links"] = str(
            {
                "source-publication": self.publication_link,
                "source-data": self.data_link,
                "other": self.other_links,
            }
        )
        row_dict["name"] = self.name
        row_dict["publication_year"] = self.publication_year
        row_dict["doi"] = self.doi

        return row_dict

    # @staticmethod
    # def __hash__(self):
    #     sha = sha512()
    #     sha.update(self.name.encode("utf-8"))
    #     return int(sha.hexdigest(), 16)

    def __str__(self):
        return (
            f"Dataset(description='{self.description}', "
            f"nconfiguration_sets={len(self.spark_row['configuration_sets'])}, "
            f"nproperty_objects={self.spark_row['nproperty_objects']}, "
            f"nconfigurations={self.spark_row['nconfigurations']}"
        )

    def __repr__(self):
        return str(self)



import pytest
import tempfile
import numpy as np
np.random.seed(42)
import random
random.seed(42)
from ase import Atoms

from colabfit import ATOMS_NAME_FIELD, ATOMS_LABELS_FIELD, ID_FORMAT_STRING
from colabfit.tools.configuration import Configuration
from colabfit.tools.database import MongoDatabase
from colabfit.tools.property_settings import PropertySettings


def build_n(n):
    images              = []
    energies            = []
    stress              = []
    names               = []
    nd_same_shape       = []
    nd_diff_shape       = []
    forces              = []
    nd_same_shape_arr   = []
    nd_diff_shape_arr   = []

    for i in range(1, n+1):
        atoms = Atoms(f'H{i}', positions=np.random.random((i, 3)))

        atoms.info['energy'] = np.random.random()
        atoms.info['stress'] = np.random.random(6)
        atoms.info['name'] = f'configuration_{i}'
        atoms.info['nd-same-shape'] = np.random.random((2, 3, 5))
        atoms.info['nd-diff-shapes'] = np.random.random((
            i+np.random.randint(1, 4),
            i+1+np.random.randint(1, 4),
            i+2+np.random.randint(1, 4),
        ))

        energies.append(atoms.info['energy'])
        stress.append(atoms.info['stress'])
        names.append(atoms.info['name'])
        nd_same_shape.append(atoms.info['nd-same-shape'])
        nd_diff_shape.append(atoms.info['nd-diff-shapes'])

        atoms.arrays['forces'] = np.random.random((i, 3))
        atoms.arrays['nd-same-shape-arr'] = np.random.random((i, 2, 3))
        atoms.arrays['nd-diff-shapes-arr'] = np.random.random((
            i,
            i+np.random.randint(1, 4),
            i+1+np.random.randint(1, 4),
        ))

        forces.append(atoms.arrays['forces'])
        nd_same_shape_arr.append(atoms.arrays['nd-same-shape-arr'])
        nd_diff_shape_arr.append(atoms.arrays['nd-diff-shapes-arr'])

        images.append(Configuration.from_ase(atoms))

    return (
        images,
        energies, stress, names, nd_same_shape, nd_diff_shape,
        forces, nd_same_shape_arr, nd_diff_shape_arr,
    )


class TestMongoDatabase:
    database_name = 'colabfit_test'

    # def test_from_html(self):
    #    with tempfile.TemporaryFile() as tmpfile:

    #         database = MongoDatabase(self.database_name, drop_database=True)

    #         # Just makes sure from_markdown doesn't throw an error
    #         dataset = database.dataset_from_markdown(
    #             'colabfit/tests/files/test.md',
    #             generator=True
    #         )

    #         assert 3 == dataset.aggregated_info['nconfigurations']
    #         assert 3 == sum(dataset.aggregated_info['property_types_counts'])


    def test_add_then_update_nochange_config(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = MongoDatabase(self.database_name, drop_database=True)

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            ids = list(database.insert_data(images))

            for co_data in database.get_data('configurations', 'names'):
                assert len(co_data) == 0

            for co_data in database.get_data('configurations', 'labels'):
                assert len(co_data) == 0

            for co_data in database.get_data('configurations', 'last_modified'):
                assert len(co_data)

            np.testing.assert_allclose(
                database.get_data('configurations', 'atomic_numbers', concatenate=True),
                np.concatenate([_.get_atomic_numbers() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations', 'positions', concatenate=True),
                np.concatenate([_.get_positions() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations', 'cell', concatenate=True),
                np.concatenate([_.get_cell() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations', 'pbc', vstack=True),
                np.vstack([_.get_pbc() for _ in images]).astype(int)
            )


    def test_add_then_update_with_changes_config(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = MongoDatabase(self.database_name, drop_database=True)

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            list(database.insert_data(images))

            for img in images:
                img.info[ATOMS_NAME_FIELD].add('change')
                img.info[ATOMS_LABELS_FIELD].add('another_label')

            list(database.insert_data(images))

            for n in database.get_data('configurations', 'names', concatenate=True):
                assert n == 'change'

            for n in database.get_data('configurations', 'labels', concatenate=True):
                assert n == 'another_label'

            for img in images:
                img.info[ATOMS_NAME_FIELD].add('change2')
                img.info[ATOMS_LABELS_FIELD] = {'another_label2'}

            ids = list(database.insert_data(images))

            for n in database.get_data('configurations', 'names'):
                assert n[0] == 'change'
                assert n[1] == 'change2'

            for n in database.get_data('configurations', 'labels'):
                assert n[0] == 'another_label'
                assert n[1] == 'another_label2'

            database.get_configuration(ids[0][0])

            np.testing.assert_allclose(
                database.get_data('configurations', 'atomic_numbers', concatenate=True),
                np.concatenate([_.get_atomic_numbers() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations', 'positions', concatenate=True),
                np.concatenate([_.get_positions() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations', 'cell', concatenate=True),
                np.concatenate([_.get_cell() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations', 'pbc', vstack=True),
                np.vstack([_.get_pbc() for _ in images]).astype(int)
            )



    def test_add_configs_props_diff_def(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = MongoDatabase(self.database_name, drop_database=True)

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'dft-energy', 'units': 'eV'},
                    'stress': {'field': 'dft-stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'dft-forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},
                }
            }

            for img in images:
                img.info['dft-energy'] = img.info['energy']
                img.info['dft-stress'] = img.info['stress']
                img.arrays['dft-forces'] = img.arrays['forces']

                del img.info['energy']
                del img.info['stress']
                del img.arrays['forces']

            list(zip(*database.insert_data(
                images, property_map=property_map
            )))

            np.testing.assert_allclose(
                database.get_data('properties', 'default.energy', ravel=True),
                np.hstack(energies)
            )
            np.testing.assert_allclose(
                database.get_data('properties', 'default.stress', ravel=True),
                np.hstack(stress)
            )
            decoded_names = database.get_data('properties', 'default.name', ravel=True)
            decoded_names = decoded_names.tolist()
            assert decoded_names == names
            np.testing.assert_allclose(
                database.get_data('properties', 'default.nd-same-shape', ravel=True),
                np.concatenate(nd_same_shape).ravel()
            )
            data = database.get_data('properties', 'default.nd-diff-shapes')
            for a1, a2 in zip(data, nd_diff_shape):
                np.testing.assert_allclose(a1, a2)

            np.testing.assert_allclose(
                database.get_data('properties', 'default.forces', concatenate=True),
                np.concatenate(forces)
            )
            np.testing.assert_allclose(
                database.get_data('properties', 'default.nd-same-shape-arr', concatenate=True),
                np.concatenate(nd_same_shape_arr)
            )
            data = database.get_data('properties', 'default.nd-diff-shapes-arr')
            for a1, a2 in zip(data, nd_diff_shape_arr):
                np.testing.assert_allclose(a1, a2)


    def test_add_then_update_with_properties_with_change(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = MongoDatabase(self.database_name, drop_database=True)

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'energy', 'units': 'eV'},
                    'stress': {'field': 'stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},

                    '_settings': {
                        '_method': 'VASP',
                        '_description': 'A basic test calculation',
                        '_files': [('dummy_name', 'dummy file contents')],
                    }
                }
            }

            database.insert_data(
                images,
                property_map=property_map,
            )

            for img in images:
                img.info['energy'] += 100000

            database.insert_data(
                images,
                property_map=property_map,
            )

            for img in images:
                img.info['energy'] += 100000

            database.insert_data(
                images,
                property_map=property_map,
            )

            np.testing.assert_allclose(
                database.get_data('configurations', 'atomic_numbers', concatenate=True),
                np.concatenate([_.get_atomic_numbers() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations', 'positions', concatenate=True),
                np.concatenate([_.get_positions() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations', 'cell', concatenate=True),
                np.concatenate([_.get_cell() for _ in images])
            )
            np.testing.assert_allclose(
                database.get_data('configurations', 'pbc', concatenate=True),
                np.concatenate([_.get_pbc() for _ in images]).astype(int)
            )

            np.testing.assert_allclose(
                database.get_data('properties', 'default.energy', ravel=True),
                np.concatenate([
                    np.hstack(energies),
                    np.hstack(energies)+100000,
                    np.hstack(energies)+200000
                ]).ravel()
            )
            np.testing.assert_allclose(
                database.get_data('properties', 'default.stress', concatenate=True),
                np.concatenate([
                    np.hstack(stress),
                    np.hstack(stress),
                    np.hstack(stress),
                ])
            )
            decoded_names = database.get_data('properties', 'default.name')
            decoded_names = decoded_names.tolist()
            assert decoded_names == names*3
            np.testing.assert_allclose(
                database.get_data('properties', 'default.nd-same-shape', concatenate=True),
                np.concatenate([
                    np.concatenate(nd_same_shape),
                    np.concatenate(nd_same_shape),
                    np.concatenate(nd_same_shape),
                ])
            )
            data = database.get_data('properties', 'default.nd-diff-shapes')
            for a1, a2 in zip(data, nd_diff_shape*3):
                np.testing.assert_allclose(a1, a2)

            np.testing.assert_allclose(
                database.get_data('properties', 'default.forces', concatenate=True),
                np.concatenate([
                    np.concatenate(forces),
                    np.concatenate(forces),
                    np.concatenate(forces),
                ])
            )
            np.testing.assert_allclose(
                database.get_data('properties', 'default.nd-same-shape-arr', concatenate=True),
                np.concatenate([
                    np.concatenate(nd_same_shape_arr),
                    np.concatenate(nd_same_shape_arr),
                    np.concatenate(nd_same_shape_arr),
                ])
            )
            data = database.get_data('properties', 'default.nd-diff-shapes-arr')
            for a1, a2 in zip(data, nd_diff_shape_arr*3):
                np.testing.assert_allclose(a1, a2)


    def test_add_then_update_with_properties_nochange(self):
        with tempfile.TemporaryFile() as tmpfile:

            database = MongoDatabase(self.database_name, drop_database=True)

            returns = build_n(10)

            images              = returns[0]
            energies            = returns[1]
            stress              = returns[2]
            names               = returns[3]
            nd_same_shape       = returns[4]
            nd_diff_shape       = returns[5]
            forces              = returns[6]
            nd_same_shape_arr   = returns[7]
            nd_diff_shape_arr   = returns[8]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'energy', 'units': 'eV'},
                    'stress': {'field': 'stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},

                    '_settings': {
                        '_method': 'VASP',
                        '_description': 'A basic test calculation',
                        '_files': [('dummy_name', 'dummy file contents')],
                    }
                }
            }

            database.insert_data(
                images,
                property_map=property_map,
            )

            database.insert_data(
                images,
                property_map=property_map,
            )

            np.testing.assert_allclose(
                database.get_data('properties', 'default.energy'),
                np.hstack(energies)
            )
            np.testing.assert_allclose(
                database.get_data('properties', 'default.stress', concatenate=True),
                np.hstack(stress)
            )
            decoded_names =  database.get_data('properties', 'default.name')
            decoded_names = decoded_names.tolist()
            assert decoded_names == names
            np.testing.assert_allclose(
                database.get_data('properties', 'default.nd-same-shape', concatenate=True),
                np.concatenate(nd_same_shape)
            )
            data = database.get_data('properties', 'default.nd-diff-shapes')
            for a1, a2 in zip(data, nd_diff_shape):
                np.testing.assert_allclose(a1, a2)

            np.testing.assert_allclose(
                database.get_data('properties', 'default.forces', concatenate=True),
                np.concatenate(forces)
            )
            np.testing.assert_allclose(
                database.get_data('properties', 'default.nd-same-shape-arr', concatenate=True),
                np.concatenate(nd_same_shape_arr)
            )
            data = database.get_data('properties', 'default.nd-diff-shapes-arr')
            for a1, a2 in zip(data, nd_diff_shape_arr):
                np.testing.assert_allclose(a1, a2)


    def test_get_configurations(self):

        with tempfile.TemporaryFile() as tmpfile:
            database = MongoDatabase(self.database_name, drop_database=True)

            images = build_n(10)[0]

            database.insert_data(images)

            count = 0
            for atoms, img in zip(database.get_configurations('all'), images):
                assert atoms == img
                count += 1
            assert count == 10


    def test_get_configurations_attach_settings(self):

        database = MongoDatabase(self.database_name, drop_database=True)

        images = build_n(10)[0]

        database.insert_property_definition(
            {
                'property-id': 'default',
                'property-title': 'A default property used for testing',
                'property-description': 'A description of the property',

                'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},

                # 'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                # 'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                # 'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                # 'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                # 'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                # 'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
            }
        )

        property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
                'forces': {'field': 'forces', 'units': 'eV/Ang'},
                'stress': {'field': 'stress', 'units': 'GPa'},

                '_settings': {

                    '_method': 'VASP',
                    '_description': 'A basic test calculation',
                    '_files': [('dummy_name', 'dummy file contents')],
                    '_labels': ['ps_label1', 'ps_label2'],

                    'name':             {'required': False, 'field': 'name', 'units': None},
                    'nd-same-shape':    {'required': False, 'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes':   {'required': False, 'field': 'nd-diff-shapes', 'units': 'eV'},
                    'nd-same-shape-arr':    {'required': False, 'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr':   {'required': False, 'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},
                }
            }
        }

        for i, img in enumerate(images):
            img.info[ATOMS_NAME_FIELD].add(f'config_{i}')
            img.info[ATOMS_LABELS_FIELD].add('a_label')

        ids = database.insert_data(
            images,
            property_map=property_map,
        )

        rebuilt_configs = database.get_configurations(
            [_[0] for _ in ids],
            attach_properties=True,
            attach_settings=True,
        )

        for i, ((cid, pid), config) in enumerate(zip(ids, images)):
            config_doc = next(database.configurations.find({'_id': cid}))
            prop_doc   = next(database.properties.find({'_id': pid}))

            pn = database.get_data(
                'properties', 'default.forces', ids=[pid], concatenate=True
            ).shape[0]

            na = len(config)
            assert config_doc['nsites'] == na
            assert pn == na

            assert config_doc['chemical_formula_anonymous'] == 'A'
            assert config_doc['chemical_formula_hill'] == config.get_chemical_formula()
            assert config_doc['chemical_formula_reduced'] == 'H'
            assert config_doc['dimension_types'] == [0, 0, 0]
            assert config_doc['elements'] == ['H']
            assert config_doc['elements_ratios'] == [1.0]
            assert {'a_label'}.issubset(config_doc['labels'])
            np.testing.assert_allclose(
                config_doc['lattice_vectors'],
                np.array(config.get_cell())
            )
            assert config_doc['names'] == [f'config_{i}']
            assert config_doc['nsites'] == len(config)
            assert config_doc['nelements'] == 1
            assert config_doc['nperiodic_dimensions'] == 0
            assert {pid}.issubset(config_doc['relationships']['properties'])

            assert {cid}.issubset(prop_doc['relationships']['configurations'])

            assert database.property_settings.count_documents({
                'relationships.properties': pid
            })


            assert rebuilt_configs[i].info['_settings._method'] == 'VASP'
            assert rebuilt_configs[i].info['_settings._description'] == 'A basic test calculation'
            assert set(rebuilt_configs[i].info['_settings._labels']) == {'ps_label1', 'ps_label2'}

        database.drop_database(database.database_name)


    def test_insert_pso_definition_data(self):

        with tempfile.NamedTemporaryFile() as tmpfile:
            database = MongoDatabase(self.database_name, drop_database=True)

            images = build_n(10)[0]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'energy', 'units': 'eV'},
                    'stress': {'field': 'stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},

                    '_settings': {
                        '_method': 'VASP',
                        '_description': 'A basic test calculation',
                        '_files': [('dummy_name', 'dummy file contents')],
                    }
                }
            }

            for i, img in enumerate(images):
                img.info[ATOMS_NAME_FIELD].add(f'config_{i}')
                img.info[ATOMS_LABELS_FIELD].add('a_label')

            ids = database.insert_data(
                images,
                property_map=property_map,
            )

            for i, ((cid, pid), config) in enumerate(zip(ids, images)):
                config_doc = next(database.configurations.find({'_id': cid}))
                prop_doc   = next(database.properties.find({'_id': pid}))

                pn = database.get_data(
                    'properties', 'default.forces', ids=[pid], concatenate=True
                ).shape[0]

                na = len(config)
                assert config_doc['nsites'] == na
                assert pn == na

                assert config_doc['chemical_formula_anonymous'] == 'A'
                assert config_doc['chemical_formula_hill'] == config.get_chemical_formula()
                assert config_doc['chemical_formula_reduced'] == 'H'
                assert config_doc['dimension_types'] == [0, 0, 0]
                assert config_doc['elements'] == ['H']
                assert config_doc['elements_ratios'] == [1.0]
                assert {'a_label'}.issubset(config_doc['labels'])
                np.testing.assert_allclose(
                    config_doc['lattice_vectors'],
                    np.array(config.get_cell())
                )
                assert config_doc['names'] == [f'config_{i}']
                assert config_doc['nsites'] == len(config)
                assert config_doc['nelements'] == 1
                assert config_doc['nperiodic_dimensions'] == 0
                assert {pid}.issubset(config_doc['relationships']['properties'])

                assert {cid}.issubset(prop_doc['relationships']['configurations'])

                assert database.property_settings.count_documents({
                    'relationships.properties': pid
                })

            database.drop_database(database.database_name)


    def test_insert_cs(self):

        with tempfile.NamedTemporaryFile() as tmpfile:
            database = MongoDatabase(self.database_name, drop_database=True)

            images = build_n(10)[0]

            for i, img in enumerate(images):
                img.info[ATOMS_NAME_FIELD].add(f'config_{i}')
                img.info[ATOMS_LABELS_FIELD].add('a_label')

            ids = database.insert_data(images)

            co_ids = list(zip(*ids))[0]

            cs_id = database.insert_configuration_set(co_ids, 'a description')

            cs_doc = next(database.configuration_sets.find({'_id': cs_id}))

            agg_info = cs_doc['aggregated_info']

            assert cs_doc['description'] == 'a description'

            assert agg_info['nconfigurations'] == len(ids)
            assert agg_info['nsites'] == sum(len(c) for c in images)
            assert agg_info['nelements'] == 1
            assert agg_info['elements'] == ['H']
            assert agg_info['individual_elements_ratios'] == {'H': [1.0]}
            assert agg_info['total_elements_ratios'] == {'H': 1.0}
            assert agg_info['labels'] == ['a_label']
            assert agg_info['labels_counts'] == [len(ids)]
            assert agg_info['chemical_formula_reduced'] == ['H']
            assert agg_info['chemical_formula_anonymous'] == ['A']
            assert set(agg_info['chemical_formula_hill']) == {
                f'H{i+1}' if i > 0 else 'H' for i in range(len(ids))
            }
            assert agg_info['nperiodic_dimensions'] == [0]
            assert agg_info['dimension_types'] == [[0,0,0]]

            database.drop_database(database.database_name)

    def test_insert_ds_diff_cs(self):

        with tempfile.NamedTemporaryFile() as tmpfile:
            database = MongoDatabase(self.database_name, drop_database=True)

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'energy', 'units': 'eV'},
                    'stress': {'field': 'stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},

                    '_settings': {
                        '_method': 'VASP',
                        '_description': 'A basic test calculation',
                        '_files': [('dummy_name', 'dummy file contents')],
                        '_labels': ['pso_label1', 'pso_label2']
                    }
                }
            }

            # pso = PropertySettings(
            #     method='VASP',
            #     description='A basic test calculation',
            #     files=[('dummy_name', 'dummy file contents')],
            #     labels=['pso_label1', 'pso_label2']
            # )

            # pso_id = database.insert_property_settings(pso)

            images = build_n(10)[0]

            for i, img in enumerate(images):
                img.info[ATOMS_NAME_FIELD].add(f'config_{i}')
                img.info[ATOMS_LABELS_FIELD].add('a_label')

            ids = database.insert_data(
                images,
                property_map=property_map,
            )

            co_ids1, pr_ids1 = list(zip(*ids))

            cs_id1 = database.insert_configuration_set(co_ids1, 'a description1')

            images = build_n(10)[0]

            for i, img in enumerate(images):
                img.info[ATOMS_NAME_FIELD].add(f'second_config_{i}')
                img.info[ATOMS_LABELS_FIELD].add('a_second_label')

                img.info['energy'] += 100000

            ids = database.insert_data(
                images,
                property_map=property_map,
            )

            co_ids2, pr_ids2 = list(zip(*ids))

            cs_id2 = database.insert_configuration_set(co_ids2, 'a description2')

            ds_id = database.insert_dataset(
                cs_ids=[cs_id1, cs_id2],
                pr_ids=pr_ids1+pr_ids2,
                name='example_dataset',
                authors=['colabfit'],
                links=['https://colabfit.org'],
                description='an example dataset',
                resync=True
            )

            ds_doc = next(database.datasets.find({'_id': ds_id}))

            assert ds_doc['authors'] == ['colabfit']
            assert ds_doc['links'] == ['https://colabfit.org']
            assert ds_doc['description'] == 'an example dataset'
            assert len(ds_doc['relationships']['configuration_sets']) == 2
            assert len(ds_doc['relationships']['properties']) == 20

            agg = ds_doc['aggregated_info']

            assert agg['nconfigurations'] == 20
            assert agg['nsites'] == 110
            assert agg['nelements'] == 1
            assert agg['elements'] == ['H']
            assert agg['individual_elements_ratios'] == {'H': [1.0]}
            assert agg['total_elements_ratios'] == {'H': 1.0}
            assert {'a_label', 'a_second_label'}.issubset(agg['configuration_labels'])
            assert agg['chemical_formula_reduced'] == ['H']
            assert agg['chemical_formula_anonymous'] == ['A']
            assert set(agg['chemical_formula_hill']) == {
                f'H{i+1}' if i > 0 else 'H' for i in range(len(ids))
            }
            assert agg['nperiodic_dimensions'] == [0]
            assert agg['dimension_types'] == [[0,0,0]]

            assert agg['property_types'] == ['default']
            assert set(agg['property_labels']) == {'pso_label1', 'pso_label2'}

            database.drop_database(database.database_name)


class TestPropertyDefinitionsAndSettings:
    database_name = 'colabfit_test'

    def test_invalid_definition(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = MongoDatabase(self.database_name, drop_database=True)

            property_definition = {
                'property-id': 'this should throw an error',
            }

            with pytest.raises(Exception):
                database.insert_property_definition(property_definition)


    def test_definition_setter_getter(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = MongoDatabase(self.database_name, drop_database=True)

            property_definition = {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shape': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }

            database.insert_property_definition(property_definition)
            
            get_def = database.get_property_definition('default')['definition']
            get_def['property-id'] = 'default'

            assert  get_def == property_definition


    def test_settings_setter_getter(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = MongoDatabase(self.database_name, drop_database=True)

            dummy_file_contents = 'this is a dummy file\nwith nonsense contents'

            pso = PropertySettings(
                method='VASP',
                description='A basic test calculation',
                files=[('dummy_name', dummy_file_contents)],
            )

            pso_id = database.insert_property_settings(pso)

            rebuilt_pso = database.get_property_settings(pso_id)

            assert pso == rebuilt_pso


    def test_settings_duplicate(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = MongoDatabase(self.database_name, drop_database=True)

            dummy_file_contents = 'this is a dummy file\nwith nonsense contents'

            pso = PropertySettings(
                method='VASP',
                description='A basic test calculation',
                files=[('dummy_name', dummy_file_contents)],
            )

            pso_id = database.insert_property_settings(pso)
            pso_id = database.insert_property_settings(pso)

            rebuilt_pso = database.get_property_settings(pso_id)

            assert pso == rebuilt_pso

class TestConfigurationSets:

    database_name = 'colabfit_test'

    def test_insert_cs(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = MongoDatabase(self.database_name, drop_database=True)

            images = build_n(10)[0]

            ids = [_[0] for _ in database.insert_data(images)]

            cs_id = database.insert_configuration_set(
                ids, description='A basic configuration set'
            )

            desc = next(database.configuration_sets.find({'_id': cs_id}))['description']

            assert desc == 'A basic configuration set'

            rebuilt_ids = next(database.configuration_sets.find({'_id': cs_id}))['relationships']['configurations']

            # rebuilt_ids = database.get_data(
            #     f'configuration_sets/{cs_id}/ids',
            #     ravel=True, in_memory=True, as_str=True
            # ).tolist()

            assert rebuilt_ids == ids

            for img in images:
                img2 = database.get_configuration(
                    ID_FORMAT_STRING.format('CO', hash(img), 0)
                )

                assert img == img2

class TestDatasets:
    database_name = 'colabfit_test'

    def test_insert_ds(self):
        with tempfile.TemporaryFile() as tmpfile:
            database = MongoDatabase(self.database_name, drop_database=True)

            images = build_n(10)[0]

            database.insert_property_definition(
                {
                    'property-id': 'default',
                    'property-title': 'A default property used for testing',
                    'property-description': 'A description of the property',
                    'energy': {'type': 'float', 'has-unit': True, 'extent': [], 'required': True, 'description': 'empty'},
                    'stress': {'type': 'float', 'has-unit': True, 'extent': [6], 'required': True, 'description': 'empty'},
                    'name': {'type': 'string', 'has-unit': False, 'extent': [], 'required': True, 'description': 'empty'},
                    'nd-same-shape': {'type': 'float', 'has-unit': True, 'extent': [2,3,5], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes': {'type': 'float', 'has-unit': True, 'extent': [":", ":", ":"], 'required': True, 'description': 'empty'},
                    'forces': {'type': 'float', 'has-unit': True, 'extent': [":", 3], 'required': True, 'description': 'empty'},
                    'nd-same-shape-arr': {'type': 'float', 'has-unit': True, 'extent': [':', 2, 3], 'required': True, 'description': 'empty'},
                    'nd-diff-shapes-arr': {'type': 'float', 'has-unit': True, 'extent': [':', ':', ':'], 'required': True, 'description': 'empty'},
                }
            )

            property_map = {
                'default': {
                    'energy': {'field': 'energy', 'units': 'eV'},
                    'stress': {'field': 'stress', 'units': 'GPa'},
                    'name': {'field': 'name', 'units': None},
                    'nd-same-shape': {'field': 'nd-same-shape', 'units': 'eV'},
                    'nd-diff-shapes': {'field': 'nd-diff-shapes', 'units': 'eV'},
                    'forces': {'field': 'forces', 'units': 'eV/Ang'},
                    'nd-same-shape-arr': {'field': 'nd-same-shape-arr', 'units': 'eV/Ang'},
                    'nd-diff-shapes-arr': {'field': 'nd-diff-shapes-arr', 'units': 'eV/Ang'},
                }
            }

            ids = database.insert_data(
                images, property_map=property_map
            )

            co_ids1, pr_ids1 = list(zip(*ids))

            cs_id1 = database.insert_configuration_set(
                co_ids1, description='A basic configuration set'
            )

            images = build_n(10)[0]

            for img in images:
                img.info['energy'] += 100000

            ids = database.insert_data(
                images, property_map=property_map
            )

            co_ids2, pr_ids2 = list(zip(*ids))

            cs_id2 = database.insert_configuration_set(
                co_ids2, description='A basic configuration set'
            )

            ds_id = database.insert_dataset(
                cs_ids=[cs_id1, cs_id2],
                pr_ids=pr_ids1+pr_ids2,
                name='example_dataset',
                authors='colabfit',
                links='https://colabfit.openkim.org/',
                description='An example dataset',
                resync=True
            )

            ds_doc = next(database.datasets.find({'_id': ds_id}))
            assert ds_doc['authors'] == ['colabfit']


from colabfit.tools import Dataset

ds1 = Dataset.from_markdown(
    '/home/josh/colabfit/data/formatted/V_PRM2019/README.md',
    convert_units=True,
    verbose=True
)

ds2 = Dataset.from_markdown(
    '/home/josh/colabfit/data/formatted/V_PRM2019/README.md',
    convert_units=True,
    verbose=True
)

# ds1.merge(ds2)
# n_dup = ds1.clean()

ds = Dataset('nested')
ds.attach_dataset(ds1)
ds.attach_dataset(ds2)

ds = ds.flatten()

print(len(ds.data))

# names = [
#     'InP_JPCA2020',
#     'MoNbTaVW_PRB2021',
#     'Mo_PRM2019',
#     'Nb_PRM2019',
#     'Ta_Linear_JCP2015',
#     'Ta_PRM2019',
#     'V_PRM2019',
#     'WBe_PRB2019',
#     'W_PRB2019',
#     'TiZrHfTa_APS2021',
#     'CuPd_CMS2019',
#     'CoNbV_CMS2019',
#     'AlNiTi_CMS2019',
# ]

# master = Dataset('master')

# for i, name in enumerate(names):
#     print('Merging', name)

#     master.merge(
#         Dataset.from_markdown(
#             f'/home/josh/colabfit/data/formatted/{name}/README.md',
#             convert_units=True,
#             verbose=True
#         ),
#         clean=False
#     )

# n_dup = master.clean()

# print(f'Discovered {n_dup} duplicate data entries')

# master.resync()

# # print(master)


import unittest
import numpy as np
from ase import Atoms

from colabfit import ATOMS_CONSTRAINTS_FIELD, ATOMS_NAME_FIELD, ATOMS_LABELS_FIELD
from colabfit.tools.configuration import Configuration


class TestConfigurations(unittest.TestCase):
    def test_empty_constructor(self):
        conf = Configuration()

        self.assertEqual(conf.info[ATOMS_NAME_FIELD], set())
        self.assertSetEqual(conf.info[ATOMS_LABELS_FIELD], set())


    def test_from_ase(self):
        atoms = Atoms('H4', pbc=True)
        atoms.info[ATOMS_NAME_FIELD] = 'test'
        labels = {'label1', 'label2'}
        atoms.info[ATOMS_LABELS_FIELD] = labels

        natoms = len(atoms)

        cell = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        pos = np.random.random((natoms, 3))

        atoms.cell = cell
        atoms.positions = pos

        conf = Configuration.from_ase(atoms)

        # self.assertEqual(conf, atoms)

        self.assertEqual(conf.info[ATOMS_NAME_FIELD], {'test'})
        self.assertSetEqual(conf.info[ATOMS_LABELS_FIELD], labels)

        np.testing.assert_allclose(conf.arrays['positions'], pos)
        np.testing.assert_allclose(conf.arrays['numbers'], np.ones(natoms))
        np.testing.assert_allclose(np.array(conf.cell), cell)
        np.testing.assert_allclose(np.array(conf.pbc), np.array([True]*3))


    def test_hashing_identical(self):
        atoms = Atoms('H4', pbc=True)
        atoms.info[ATOMS_NAME_FIELD] = 'test'
        labels = {'label1', 'label2'}
        atoms.info[ATOMS_LABELS_FIELD] = labels

        natoms = len(atoms)

        cell = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        pos = np.random.random((natoms, 3))

        atoms.cell = cell
        atoms.positions = pos

        conf1 = Configuration.from_ase(atoms)
        conf2 = Configuration.from_ase(atoms)

        self.assertEqual(hash(conf1), hash(conf2))


    def test_hashing_different(self):
        atoms = Atoms('H4', pbc=True)
        atoms.info[ATOMS_NAME_FIELD] = 'test'
        labels = {'label1', 'label2'}
        atoms.info[ATOMS_LABELS_FIELD] = labels

        natoms = len(atoms)

        cell = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        pos = np.random.random((natoms, 3))

        atoms.cell = cell
        atoms.positions = pos

        atoms2 = Atoms(atoms)
        atoms2.positions += 1

        atoms3 = Atoms(atoms2)
        atoms3.positions = atoms.positions

        conf1 = Configuration.from_ase(atoms)
        conf2 = Configuration.from_ase(atoms2)
        conf3 = Configuration.from_ase(atoms3)

        self.assertNotEqual(hash(conf1), hash(conf2))
        self.assertEqual(hash(conf1), hash(conf3))


    # def test_constrained_hashing_same(self):
    #     atoms = Atoms('H4', pbc=True)
    #     atoms.info[ATOMS_NAME_FIELD] = 'test'

    #     labels = {'label1', 'label2'}
    #     atoms.info[ATOMS_LABELS_FIELD] = labels

    #     atoms.info[ATOMS_CONSTRAINTS_FIELD] = {ATOMS_LABELS_FIELD}

    #     natoms = len(atoms)

    #     cell = np.array([
    #         [1, 0, 0],
    #         [0, 1, 0],
    #         [0, 0, 1]
    #     ])
    #     pos = np.random.random((natoms, 3))

    #     atoms.cell = cell
    #     atoms.positions = pos

    #     conf1 = Configuration.from_ase(atoms)
    #     conf2 = Configuration.from_ase(atoms)

    #     self.assertEqual(hash(conf1), hash(conf2))


    # def test_constrained_hashing_diff(self):
    #     atoms = Atoms('H4', pbc=True)
    #     atoms.info[ATOMS_NAME_FIELD] = 'test'
    #     labels = {'label1', 'label2'}
    #     atoms.info[ATOMS_LABELS_FIELD] = labels

    #     atoms.info[ATOMS_CONSTRAINTS_FIELD] = {ATOMS_LABELS_FIELD}

    #     natoms = len(atoms)

    #     cell = np.array([
    #         [1, 0, 0],
    #         [0, 1, 0],
    #         [0, 0, 1]
    #     ])
    #     pos = np.random.random((natoms, 3))

    #     atoms.cell = cell
    #     atoms.positions = pos

    #     atoms2 = Atoms(atoms)

    #     atoms2.info[ATOMS_LABELS_FIELD] = {'a_different_label'}

    #     conf1 = Configuration.from_ase(atoms)
    #     conf2 = Configuration.from_ase(atoms2)

    #     self.assertNotEqual(hash(conf1), hash(conf2))


import numpy as np
from ase import Atoms

from colabfit.tools.database import MongoDatabase
from colabfit.tools.configuration import Configuration

def main():
    database = MongoDatabase('debug', drop_database=True)

    configurations = []
    for i in range(1, 101):
        atoms = Atoms(
            symbols=f'H{i}',
            positions=np.random.random((i, 3)),
            cell=[[1,0,0],[0,1,0],[0,0,1]],
        )

        atoms.info['property-data.data'] = np.random.random((100, 1000))

        configurations.append(Configuration.from_ase(atoms))

    property_map = {
        'property-data': [{
            'data': {'field': 'property-data.data', 'units': None}
        }]
    }

    database.insert_property_definition({
        'property-id': 'property-data',
        'property-title': 'A title',
        'property-description': 'A description',

        'data': {
            'type': 'float',
            'has-unit': False,
            'extent': [100,1000],
            'required': True,
            'description': 'Some data'
        }
    })

    ids = database.insert_data(
        configurations,
        property_map,
        generator=True
    )

    co_ids, pr_ids = list(zip(*ids))

    rebuilt_configs = database.get_configurations(
        co_ids, pr_ids, attach_properties=True
    )

    data = database.get_data('properties', '_id')


if __name__ == '__main__':
    main()


import re
import unittest
import numpy as np
from ase import Atoms

from colabfit import ATOMS_NAME_FIELD, ATOMS_LABELS_FIELD
from colabfit.tools.configuration import Configuration
from colabfit.tools.dataset import Dataset, load_data
from colabfit.tools.property import MissingPropertyFieldWarning



class TestDatasetConstruction(unittest.TestCase):
    def test_from_html(self):
        # Just makes sure from_markdown doesn't throw an error
        dataset = Dataset.from_markdown('colabfit/tests/files/test.md')

        self.assertEqual(6, len(dataset.data))
        self.assertEqual(3, len(dataset.get_data('a-custom-field-name')))
        self.assertEqual(3, len(dataset.get_data('energy')))


    def test_config_setter(self):
        atoms = []
        for i in range(3):
            atoms.append(Atoms())
            atoms[-1].info[ATOMS_NAME_FIELD] = i

        dataset = Dataset(name='test')
        dataset.configurations = [Configuration.from_ase(at) for at in atoms]

        self.assertEqual(
            ['0', '1', '2'],
            [c.info[ATOMS_NAME_FIELD] for c in dataset.configurations]
        )

        del dataset.configurations[1]

        self.assertEqual(
            ['0', '2'],
            [c.info[ATOMS_NAME_FIELD] for c in dataset.configurations]
        )

        new_atoms = Atoms()
        new_atoms.info[ATOMS_NAME_FIELD] = 4
        dataset.configurations += [Configuration.from_ase(new_atoms)]

        self.assertEqual(
            ['0', '2', '4'],
            [c.info[ATOMS_NAME_FIELD] for c in dataset.configurations]
        )

    def test_co_label_refresher(self):

        dataset = Dataset(name='test')
        atoms = []
        for ii in range(10):
            atoms.append(Atoms())
            atoms[-1].info[ATOMS_NAME_FIELD] = ii

        dataset.configurations = [Configuration.from_ase(at) for at in atoms]

        dataset.configuration_label_regexes = {
            '[0-4]': {'0_to_4'},
            '[5-9]': {'5_to_9'},
        }

        dataset.refresh_config_labels()

        # Make sure label refresh is working
        for ai, conf in enumerate(dataset.configurations):
            self.assertEqual(conf.info[ATOMS_NAME_FIELD], f"{ai}")
            if ai < 5:
                self.assertSetEqual(conf.info[ATOMS_LABELS_FIELD], {'0_to_4'})
            elif ai >= 5:
                self.assertSetEqual(conf.info[ATOMS_LABELS_FIELD], {'5_to_9'})


        # Make sure that a new CO gets its labels updated
        new_atoms = Atoms()
        new_atoms.info[ATOMS_NAME_FIELD] = '4'
        dataset.configurations += [Configuration.from_ase(new_atoms)]

        self.assertSetEqual(
            dataset.configurations[-1].info[ATOMS_LABELS_FIELD], set()
        )

        dataset.refresh_config_labels()

        self.assertSetEqual(
            dataset.configurations[-1].info[ATOMS_LABELS_FIELD], {'0_to_4'}
        )


    def test_cs_refresher(self):

        dataset = Dataset(name='test')
        atoms = []
        for ii in range(10):
            atoms.append(Atoms())
            atoms[-1].info[ATOMS_NAME_FIELD] = ii

        dataset.configurations = [Configuration.from_ase(at)for at in atoms]

        # dataset.cs_regexes = {}

        for cs in dataset.configuration_sets:
            for conf in cs.configurations:
                if int(conf.info[ATOMS_NAME_FIELD][-1]) < 5:
                    self.assertEqual(cs.description, 'Default configuration set')
                else:
                    self.assertEqual(cs.description, 'Default configuration set')

        dataset.configuration_set_regexes = {
            'default': 'default',
            '[0-4]': '0_to_4',
            '[5-9]': '5_to_9',
        }

        for cs in dataset.configuration_sets:
            for conf in cs.configurations:
                if int(conf.info[ATOMS_NAME_FIELD][-1]) < 5:
                    self.assertEqual(cs.description, '0_to_4')
                else:
                    self.assertEqual(cs.description, '5_to_9')

        dataset.resync()
        self.assertEqual(len(dataset.configuration_sets), 2)

        dataset.configuration_set_regexes['[3-7]'] = '3_to_7'

        dataset.resync()
        self.assertEqual(len(dataset.configuration_sets), 3)


    def test_labels_updated_everywhere(self):

        dataset = Dataset(name='test')
        atoms = []
        for ii in range(10):
            atoms.append(Atoms())
            atoms[-1].info[ATOMS_NAME_FIELD] = ii
        dataset.configurations = [Configuration.from_ase(at)for at in atoms]

        dataset.configuration_set_regexes = {
            'default': 'default',
            '[0-4]': '0_to_4',
            '[5-9]': '5_to_9',
        }

        # Ensure labels are initially empty
        for cs in dataset.configuration_sets:
            for conf in cs.configurations:
                self.assertSetEqual(set(), conf.info[ATOMS_LABELS_FIELD])

        # Make sure they're re-written correctly
        dataset.configuration_label_regexes = {
            '[0-4]': {'0_to_4'},
            '[5-9]': {'5_to_9'},
        }

        for cs in dataset.configuration_sets:
            if cs.description == '0_to_4':
                self.assertSetEqual({'0_to_4'}, set(cs.labels))
            elif cs.description == '5_to_9':
                self.assertSetEqual({'5_to_9'}, set(cs.labels))

            for conf in cs.configurations:
                if int(conf.info[ATOMS_NAME_FIELD][-1]) < 5:
                    self.assertSetEqual(
                        {'0_to_4'}, conf.info[ATOMS_LABELS_FIELD]
                    )
                else:
                    self.assertSetEqual(
                        {'5_to_9'}, conf.info[ATOMS_LABELS_FIELD]
                    )

        # Make sure they're added to correctly
        dataset.configuration_label_regexes['[0-9]'] = {'new_label'}
        dataset.resync()

        for cs in dataset.configuration_sets:
            for conf in cs.configurations:
                if int(conf.info[ATOMS_NAME_FIELD][-1]) < 5:
                    self.assertSetEqual(
                        {'0_to_4', 'new_label'}, conf.info[ATOMS_LABELS_FIELD]
                    )
                else:
                    self.assertSetEqual(
                        {'5_to_9', 'new_label'}, conf.info[ATOMS_LABELS_FIELD]
                    )

        # And also removed properly
        dataset.delete_config_label_regex('[0-4]')
        dataset.resync()

        for cs in dataset.configuration_sets:
            for conf in cs.configurations:
                if int(conf.info[ATOMS_NAME_FIELD][-1]) < 5:
                    self.assertSetEqual(
                        {'new_label'}, conf.info[ATOMS_LABELS_FIELD]
                    )
                else:
                    self.assertSetEqual(
                        {'5_to_9', 'new_label'}, conf.info[ATOMS_LABELS_FIELD]
                    )

    def test_default_name(self):
        dataset = Dataset('test')

        dataset.configurations = load_data(
            'colabfit/tests/files/test_file.extxyz',
            file_format='xyz',
            name_field=None,
            elements=['In', 'P'],
            default_name='test',
        )

        self.assertEqual(
            ['test_0', 'test_1', 'test_2'],
            [co.info[ATOMS_NAME_FIELD] for co in dataset.configurations]
        )


    def test_custom_properties(self):

        dataset = Dataset(name='test')
        atoms = []
        for ii in range(1, 11):
            atoms.append(Atoms(f'H{ii}', positions=np.random.random((ii, 3))))
            atoms[-1].info[ATOMS_NAME_FIELD] = ii

            atoms[-1].info['eng'] = ii
            atoms[-1].info['fcs'] = np.ones((ii, 3))*ii

            atoms[-1].info['string'] = f'string_{ii}'
            atoms[-1].info['1d-array'] = np.ones(5)*ii
            atoms[-1].arrays['per-atom-array'] = np.ones((ii, 3))*ii+1

        dataset.configurations = [Configuration.from_ase(at)for at in atoms]

        dataset.property_map = {
            'default': {
                'energy': {'field': 'eng', 'units': 'eV'},
                'forces': {'field': 'fcs', 'units': 'eV/Ang'},
            },
            'my-custom-property': {
                'a-custom-string':
                    {'field': 'string', 'units': None},
                'a-custom-1d-array':
                    {'field': '1d-array', 'units': 'eV'},
                'a-custom-per-atom-array':
                    {'field': 'per-atom-array', 'units': 'eV'},
            }
        }

        dataset.custom_definitions = {
            'my-custom-property': 'colabfit/tests/files/test_property.edn'
        }

        dataset.parse_data()

        self.assertEqual(20, len(dataset.data))
        self.assertEqual(10, len(dataset.get_data('energy')))
        self.assertEqual(10, len(dataset.get_data('forces')))
        self.assertEqual(10, len(dataset.get_data('a-custom-string')))
        self.assertEqual(10, len(dataset.get_data('a-custom-1d-array')))
        self.assertEqual(10, len(dataset.get_data('a-custom-per-atom-array')))



    def test_custom_properties_no_edn(self):

        dataset = Dataset(name='test')
        atoms = []
        for ii in range(1, 11):
            atoms.append(Atoms(f'H{ii}', positions=np.random.random((ii, 3))))
            atoms[-1].info[ATOMS_NAME_FIELD] = ii

            atoms[-1].info['eng'] = ii
            atoms[-1].info['fcs'] = np.ones((ii, 3))*ii

            atoms[-1].info['string'] = f'string_{ii}'
            atoms[-1].info['1d-array'] = np.ones(5)*ii
            atoms[-1].arrays['per-atom-array'] = np.ones((ii, 3))*ii+1

        dataset.configurations = [Configuration.from_ase(at)for at in atoms]

        dataset.property_map = {
            'default': {
                'energy': {'field': 'eng', 'units': 'eV'},
                'forces': {'field': 'fcs', 'units': 'eV/Ang'},
            },
            'my-custom-property': {
                'a-custom-string':
                    {'field': 'string', 'units': None},
                'a-custom-1d-array':
                    {'field': '1d-array', 'units': 'eV'},
                'a-custom-per-atom-array':
                    {'field': 'per-atom-array', 'units': 'eV'},
            }
        }

        dataset.custom_definitions = {
            'my-custom-property': {
                "property-id":
                    "my-custom-property",
                "property-title":
                    "A custom, user-provided Property Definition. See "\
                    "https://openkim.org/doc/schema/properties-framework/ for "\
                    "instructions on how to build these files",
                "property-description":
                    "Some human-readable description",
                "a-custom-field-name": {
                    "type":         "string",
                    "has-unit":     False,
                    "extent":       [],
                    "required":     False,
                    "description":  "The description of the custom field"
                },
                "a-custom-1d-array": {
                    "type":         "float",
                    "has-unit":     True,
                    "extent":       [":"],
                    "required":     True,
                    "description":  "This should be a 1D vector of floats"
                },
                "a-custom-per-atom-array": {
                    "type":         "float",
                    "has-unit":     True,
                    "extent":       [":", 3],
                    "required":     True,
                    "description":
                        "This is a 2D array of floats, where the second "\
                        "dimension has a length of 3"
                },
            }
        }

        dataset.parse_data()

        self.assertEqual(20, len(dataset.data))
        self.assertEqual(10, len(dataset.get_data('energy')))
        self.assertEqual(10, len(dataset.get_data('forces')))
        self.assertEqual(10, len(dataset.get_data('a-custom-string')))
        self.assertEqual(10, len(dataset.get_data('a-custom-1d-array')))
        self.assertEqual(10, len(dataset.get_data('a-custom-per-atom-array')))


    def test_custom_properties_missing_key(self):

        dataset = Dataset(name='test')
        atoms = []
        for ii in range(1, 11):
            atoms.append(Atoms(f'H{ii}', positions=np.random.random((ii, 3))))
            atoms[-1].info[ATOMS_NAME_FIELD] = ii

            atoms[-1].info['eng'] = ii
            atoms[-1].info['fcs'] = np.ones((ii, 3))*ii

            atoms[-1].info['string'] = f'string_{ii}'
            atoms[-1].info['1d-array'] = np.ones(5)*ii

            # Remove one of the important keys
            # atoms[-1].arrays['per-atom-array'] = np.ones((ii, 3))*ii+1

        dataset.configurations = [Configuration.from_ase(at)for at in atoms]

        dataset.property_map = {
            'default': {
                'energy': {'field': 'eng', 'units': 'eV'},
                'forces': {'field': 'fcs', 'units': 'eV/Ang'},
            },
            'my-custom-property': {
                'a-custom-string':
                    {'field': 'string', 'units': None},
                'a-custom-1d-array':
                    {'field': '1d-array', 'units': 'eV'},
                'a-custom-per-atom-array':
                    {'field': 'per-atom-array', 'units': 'eV'},
            }
        }

        dataset.custom_definitions = {
            'my-custom-property': 'colabfit/tests/files/test_property.edn'
        }

        self.assertWarns(MissingPropertyFieldWarning, dataset.parse_data)


class TestSetOperations(unittest.TestCase):

    def test_subset_is_subset(self):
        dataset1 = Dataset('test1')

        images1 = []
        for i in range(5):
            images1.append(Atoms('H2', positions=np.random.random((2, 3))))
            images1[-1].info[ATOMS_NAME_FIELD] = dataset1.name + str(i)
            images1[-1].info[ATOMS_LABELS_FIELD] = dataset1.name + '_label_'+ str(i)

            images1[-1].info['energy'] = float(i)

        dataset1.configurations = [Configuration.from_ase(at) for at in images1]
        dataset1.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset1.parse_data()

        dataset2 = Dataset('test1')
        dataset2.configurations = [Configuration.from_ase(at) for at in images1[:3]]
        dataset2.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset2.parse_data()

        self.assertFalse(dataset1.issubset(dataset2))
        self.assertTrue(dataset2.issubset(dataset1))


    def test_superset_is_subset(self):
        dataset1 = Dataset('test1')

        images1 = []
        for i in range(5):
            images1.append(Atoms('H2', positions=np.random.random((2, 3))))
            images1[-1].info[ATOMS_NAME_FIELD] = dataset1.name + str(i)
            images1[-1].info[ATOMS_LABELS_FIELD] = dataset1.name + '_label_'+ str(i)

            images1[-1].info['energy'] = float(i)

        dataset1.configurations = [Configuration.from_ase(at) for at in images1]
        dataset1.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset1.parse_data()

        dataset2 = Dataset('test1')
        dataset2.configurations = [Configuration.from_ase(at) for at in images1[:3]]
        dataset2.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset2.parse_data()

        self.assertFalse(dataset2.issuperset(dataset1))
        self.assertTrue(dataset1.issuperset(dataset2))


    def test_disjoint(self):
        dataset1 = Dataset('test1')

        images1 = []
        for i in range(5):
            images1.append(Atoms('H2', positions=np.random.random((2, 3))))
            images1[-1].info[ATOMS_NAME_FIELD] = dataset1.name + str(i)
            images1[-1].info[ATOMS_LABELS_FIELD] = dataset1.name + '_label_'+ str(i)

            images1[-1].info['energy'] = float(i)

        dataset1.configurations = [Configuration.from_ase(at) for at in images1]
        dataset1.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset1.parse_data()

        dataset2 = Dataset('test1')

        images2 = []
        for i in range(5):
            images2.append(Atoms('H2', positions=np.random.random((2, 3))))
            images2[-1].info[ATOMS_NAME_FIELD] = dataset2.name + str(i)
            images2[-1].info[ATOMS_LABELS_FIELD] = dataset2.name + '_label_'+ str(i)

            images2[-1].info['energy'] = float(i) + 20

        dataset2.configurations = [Configuration.from_ase(at) for at in images2]
        dataset2.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset2.parse_data()

        self.assertFalse(dataset1.issubset(dataset2))
        self.assertFalse(dataset2.issubset(dataset1))


    def test_parent_child(self):
        dataset1 = Dataset('test1')

        images1 = []
        for i in range(5):
            images1.append(Atoms('H2', positions=np.random.random((2, 3))))
            images1[-1].info[ATOMS_NAME_FIELD] = dataset1.name + str(i)
            images1[-1].info[ATOMS_LABELS_FIELD] = dataset1.name + '_label_'+ str(i)

            images1[-1].info['energy'] = float(i)

        dataset1.configurations = [Configuration.from_ase(at) for at in images1]
        dataset1.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset1.parse_data()

        dataset2 = Dataset('test1')

        images2 = []
        for i in range(5):
            images2.append(Atoms('H2', positions=np.random.random((2, 3))))
            images2[-1].info[ATOMS_NAME_FIELD] = dataset2.name + str(i)
            images2[-1].info[ATOMS_LABELS_FIELD] = dataset2.name + '_label_'+ str(i)

            images2[-1].info['energy'] = float(i) + 20

        dataset2.configurations = [Configuration.from_ase(at) for at in images2]
        dataset2.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset2.parse_data()

        parent = Dataset('parent')
        parent.attach_dataset(dataset1)

        self.assertTrue(dataset1.issubset(parent))
        self.assertTrue(parent.issubset(dataset1))

        self.assertFalse(dataset2.issubset(parent))
        self.assertFalse(parent.issubset(dataset2))

        parent.attach_dataset(dataset2)

        self.assertTrue(dataset1.issubset(parent))
        self.assertTrue(dataset2.issubset(parent))
        self.assertFalse(parent.issubset(dataset1))
        self.assertFalse(parent.issubset(dataset2))


    def test_parent_parent(self):
        dataset1 = Dataset('test1')

        images1 = []
        for i in range(5):
            images1.append(Atoms('H2', positions=np.random.random((2, 3))))
            images1[-1].info[ATOMS_NAME_FIELD] = dataset1.name + str(i)
            images1[-1].info[ATOMS_LABELS_FIELD] = dataset1.name + '_label_'+ str(i)

            images1[-1].info['energy'] = float(i)

        dataset1.configurations = [Configuration.from_ase(at) for at in images1]
        dataset1.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset1.parse_data()

        dataset2 = Dataset('test1')

        images2 = []
        for i in range(5):
            images2.append(Atoms('H2', positions=np.random.random((2, 3))))
            images2[-1].info[ATOMS_NAME_FIELD] = dataset2.name + str(i)
            images2[-1].info[ATOMS_LABELS_FIELD] = dataset2.name + '_label_'+ str(i)

            images2[-1].info['energy'] = float(i) + 20

        dataset2.configurations = [Configuration.from_ase(at) for at in images2]
        dataset2.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset2.parse_data()

        parent1 = Dataset('parent')
        parent1.attach_dataset(dataset1)

        parent2 = Dataset('parent')
        parent2.attach_dataset(dataset1)
        parent2.attach_dataset(dataset2)

        self.assertTrue(parent1.issubset(parent2))
        self.assertFalse(parent2.issubset(parent1))


    def test_equality(self):
        dataset1 = Dataset('test1')

        images1 = []
        for i in range(5):
            images1.append(Atoms('H2', positions=np.random.random((2, 3))))
            images1[-1].info[ATOMS_NAME_FIELD] = dataset1.name + str(i)
            images1[-1].info[ATOMS_LABELS_FIELD] = dataset1.name + '_label_'+ str(i)

            images1[-1].info['energy'] = float(i)

        dataset1.configurations = [Configuration.from_ase(at) for at in images1]
        dataset1.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset1.parse_data()

        dataset2 = Dataset('test1')

        images2 = []
        for i in range(5):
            images2.append(Atoms('H2', positions=np.random.random((2, 3))))
            images2[-1].info[ATOMS_NAME_FIELD] = dataset2.name + str(i)
            images2[-1].info[ATOMS_LABELS_FIELD] = dataset2.name + '_label_'+ str(i)

            images2[-1].info['energy'] = float(i) + 20

        dataset2.configurations = [Configuration.from_ase(at) for at in images2]
        dataset2.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }
        dataset2.parse_data()

        self.assertEqual(dataset1, dataset1)
        self.assertNotEqual(dataset1, dataset2)


class TestFilter(unittest.TestCase):

    def test_filter_on_co_names(self):
        dataset = Dataset('test')

        images = []
        for i in range(5):
            images.append(Atoms('H2', positions=np.random.random((2, 3))))
            images[-1].info[ATOMS_NAME_FIELD] = dataset.name + str(i)
            images[-1].info[ATOMS_LABELS_FIELD] = dataset.name + '_label_'+ str(i)

            images[-1].info['energy'] = float(i)

        dataset.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }

        dataset.configurations = [Configuration.from_ase(at) for at in images]

        regex = re.compile('test[0-3]')
        filtered = dataset.filter(
            'configurations',
            lambda c: regex.search(c.info[ATOMS_NAME_FIELD])
        )

        self.assertEqual(len(filtered.configurations), 4)

        dataset.parse_data()

        regex = re.compile('test[0-3]')
        filtered = dataset.filter(
            'configurations',
            lambda c: regex.search(c.info[ATOMS_NAME_FIELD])
        )

        self.assertEqual(len(filtered.configurations), 4)
        self.assertEqual(len(filtered.data), 4)

        regex = re.compile('.*')
        filtered = dataset.filter(
            'configurations',
            lambda c: regex.search(c.info[ATOMS_NAME_FIELD])
        )

        self.assertEqual(dataset, filtered)


    def test_filter_parent_on_co_names(self):
        dataset = Dataset('test')

        images = []
        for i in range(5):
            images.append(Atoms('H2', positions=np.random.random((2, 3))))
            images[-1].info[ATOMS_NAME_FIELD] = dataset.name + str(i)
            images[-1].info[ATOMS_LABELS_FIELD] = dataset.name + '_label_'+ str(i)

            images[-1].info['energy'] = float(i)

        dataset.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }


        dataset.configurations = [Configuration.from_ase(at) for at in images]

        dataset.parse_data()

        regex = re.compile('test[0-2]')
        filtered1 = dataset.filter(
            'configurations',
            lambda c: regex.search(c.info[ATOMS_NAME_FIELD])
        )

        regex = re.compile('test[3-4]')
        filtered2 = dataset.filter(
            'configurations',
            lambda c: regex.search(c.info[ATOMS_NAME_FIELD])
        )

        parent = Dataset('parent')
        parent.attach_dataset(filtered1)
        parent.attach_dataset(filtered2)
        parent.resync()

        regex = re.compile('test[2-3]')
        parent2 = parent.filter(
            'configurations',
            lambda c: regex.search(c.info[ATOMS_NAME_FIELD])
        )

        self.assertEqual(parent2.data[0].configurations[0].info[ATOMS_NAME_FIELD], 'test2')
        self.assertEqual(parent2.data[1].configurations[0].info[ATOMS_NAME_FIELD], 'test3')


    def test_filter_on_data(self):
        dataset = Dataset('test')

        images = []
        for i in range(5):
            images.append(Atoms('H2', positions=np.random.random((2, 3))))
            images[-1].info[ATOMS_NAME_FIELD] = dataset.name + str(i)
            images[-1].info[ATOMS_LABELS_FIELD] = dataset.name + '_label_'+ str(i)

            images[-1].info['energy'] = float(i)

        dataset.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }


        dataset.configurations = [Configuration.from_ase(at) for at in images]

        dataset.parse_data()

        filtered = dataset.filter(
            'data',
            lambda d: d.edn['unrelaxed-potential-energy']['source-value'] < 3.0
        )

        self.assertEqual(len(filtered.configurations), 3)


    def test_filter_parent_on_data(self):
        dataset = Dataset('test')

        images = []
        for i in range(5):
            images.append(Atoms('H2', positions=np.random.random((2, 3))))
            images[-1].info[ATOMS_NAME_FIELD] = dataset.name + str(i)
            images[-1].info[ATOMS_LABELS_FIELD] = dataset.name + '_label_'+ str(i)

            images[-1].info['energy'] = float(i)

        dataset.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }


        dataset.configurations = [Configuration.from_ase(at) for at in images]

        dataset.parse_data()

        filtered1 = dataset.filter(
            'data',
            lambda d:
                0 <= d.edn['unrelaxed-potential-energy']['source-value'] <= 2
        )

        filtered2 = dataset.filter(
            'data',
            lambda d:
                3 <= d.edn['unrelaxed-potential-energy']['source-value'] <= 4
        )


        parent = Dataset('parent')
        parent.attach_dataset(filtered1)
        parent.attach_dataset(filtered2)
        parent.resync()

        parent2 = parent.filter(
            'data',
            lambda d:
                2 <= d.edn['unrelaxed-potential-energy']['source-value'] <= 3
        )


        self.assertEqual(parent2.data[0].configurations[0].info[ATOMS_NAME_FIELD], 'test2')
        self.assertEqual(parent2.data[1].configurations[0].info[ATOMS_NAME_FIELD], 'test3')


    def test_multi_layer_parent(self):
        dataset = Dataset('test')

        images = []
        for i in range(30):
            images.append(Atoms('H2', positions=np.random.random((2, 3))))
            images[-1].info[ATOMS_NAME_FIELD] = dataset.name + str(i)
            images[-1].info[ATOMS_LABELS_FIELD] = dataset.name + '_label_'+ str(i)

            images[-1].info['energy'] = float(i)

        dataset.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
            },
        }


        dataset.configurations = [Configuration.from_ase(at) for at in images]

        dataset.parse_data()

        child1 = dataset.filter(
            'data',
            lambda d:
                0 <= d.edn['unrelaxed-potential-energy']['source-value'] < 10
        )

        child2 = dataset.filter(
            'data',
            lambda d:
                0 <= d.edn['unrelaxed-potential-energy']['source-value'] < 10
        )

        child3 = dataset.filter(
            'data',
            lambda d:
                0 <= d.edn['unrelaxed-potential-energy']['source-value'] < 10
        )

        subparent = Dataset('subparent')
        subparent.attach_dataset(child1)
        subparent.attach_dataset(child2)
        subparent.resync()

        parent = Dataset('parent')
        parent.attach_dataset(subparent)
        parent.attach_dataset(child3)
        parent.resync()

        self.assertEqual(len(parent.configurations), 30)

        subset = parent.filter(
            'data',
            lambda d:
                d.edn['unrelaxed-potential-energy']['source-value'] % 5 == 0
        )

        self.assertEquals(len(subset.data[0].configurations), 4)
        self.assertEquals(len(subset.data[1].configurations), 2)


class Test_ParentDatasets(unittest.TestCase):
    def setUp(self):
        dataset = Dataset('test')

        dataset.configurations = load_data(
            'colabfit/tests/files/test_file.extxyz',
            file_format='xyz',
            name_field='name',
            elements=['In', 'P'],
            default_name='test',
        )

        dataset.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
                'forces': {'field': 'forces', 'units': 'eV/Ang'},
            }
        }

        dataset.parse_data()

        dataset.configuration_set_regexes = {
            'd0lda[7|5]': 'first two',
            'd0lda1': 'last one',
        }

        dataset.resync()
        self.dataset = dataset


    def test_basic(self):
        child0 = self.dataset.dataset_from_config_sets(0)
        child1 = self.dataset.dataset_from_config_sets(1)

        parent1 = Dataset('parent1')
        parent1.attach_dataset(child0)
        parent1.attach_dataset(child1)
        parent1.resync()

        parent2 = parent1.dataset_from_config_sets([(0, 0), (1, 0)])

        parent2.resync()

        self.assertEqual(parent1, parent2)

        eng1 = parent1.get_data('energy', ravel=True)
        eng2 = parent2.get_data('energy', ravel=True)

        np.testing.assert_allclose(eng1, eng2)


class Test_DatasetFunctions(unittest.TestCase):
    def setUp(self):
        dataset1 = Dataset('dataset1')

        images1 = []
        for i in range(5):
            images1.append(Atoms('H2', positions=np.random.random((2, 3))))
            images1[-1].info[ATOMS_NAME_FIELD] = dataset1.name + str(i)
            images1[-1].info[ATOMS_LABELS_FIELD] = dataset1.name + '_label_'+ str(i)

            images1[-1].info['energy'] = float(i)
            images1[-1].arrays['forces'] = np.ones_like(images1[-1].positions)*float(i)

        dataset1.configurations = [Configuration.from_ase(at) for at in images1]
        dataset1.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
                'forces': {'field': 'forces', 'units': 'eV/Ang'},
            }
        }

        dataset1.parse_data()
        dataset1.resync()

        dataset2 = Dataset('dataset2')

        images2 = []
        for i in range(5):
            images2.append(Atoms('H2', positions=np.random.random((2, 3))))
            images2[-1].info[ATOMS_NAME_FIELD] = dataset2.name + str(i)
            images2[-1].info[ATOMS_LABELS_FIELD] = dataset2.name + '_label_'+ str(i)

            images2[-1].info['energy'] = float(i)
            images2[-1].arrays['forces'] = np.ones_like(images2[-1].positions)*float(i)

        dataset2.configurations = [Configuration.from_ase(at) for at in images2]
        dataset2.property_map = {
            'default': {
                'energy': {'field': 'energy', 'units': 'eV'},
                'forces': {'field': 'forces', 'units': 'eV/Ang'},
            }
        }

        dataset2.parse_data()
        dataset2.resync()

        parent = Dataset('parent')
        parent.attach_dataset(dataset1)
        parent.attach_dataset(dataset2)
        parent.resync()

        self.dataset1 = dataset1
        self.dataset2 = dataset2
        self.parent   = parent


    def test_rename_property(self):
        eng_org = self.dataset1.get_data('energy', ravel=True)

        self.dataset1.rename_property('energy', 'a_different_name')

        eng_new = self.dataset1.get_data('a_different_name', ravel=True)

        np.testing.assert_allclose(eng_org, eng_new)


    def test_basic_transform(self):
        self.dataset1.apply_transformation('energy', lambda x, c: 1.0)
        eng_org = self.dataset1.get_data('energy', ravel=True)